import numpy as np
from scipy.special import gammaln, comb
from scipy.stats import norm
import math

def stable_logsumexp(x):
    a = np.max(x)
    return a+np.log(np.sum(np.exp(x-a)))

def stable_logsumexp_two(x,y):
    a = np.maximum(x,y)
    #if np.isneginf(a):
    #    return a
    #else:
    return a + np.log(np.exp(x-a) + np.exp(y-a))

def stable_log_diff_exp(x, y):
    # ensure that y > x
    # this function returns the stable version of log(exp(y)-exp(x)) if y > x
    if y > x:
        s = True
        mag = y + np.log(1 - np.exp(x - y))
    elif y < x:
        s = False
        mag = x + np.log(1 - np.exp(y - x))
    else:
        s = True
        mag = -np.inf

    return s, mag

def stable_log_sinh(x):
    assert(x >= 0)
    s, mag = stable_log_diff_exp(x, -x)
    return np.log(0.5) + mag

def _log1mexp(x):
    """ from pate Numerically stable computation of log(1-exp(x))."""
    if x < -1:
        return math.log1p(-math.exp(x))
    elif x < 0:
        return math.log(-math.expm1(x))
    elif x == 0:
        return -np.inf
    else:
        raise ValueError("Argument must be non-positive.")


def stable_sum_signed(xs, x, ys, y):
    # x and y are log abs,  xs,ys are signs
    if xs == ys:
        s = ys
        out = stable_logsumexp_two(x, y)
    elif ys:
        s, out = stable_log_diff_exp(x, y)
    else:
        s, out = stable_log_diff_exp(y, x)
    return s, out

def stable_inplace_diff_in_log(vec, signs, n=-1):
    """ This function replaces the first n-1 dimension of vec with the log of abs difference operator
     Input:
        - `vec` is a numpy array of floats with size larger than 'n'
        - `signs` is a numpy array of bools with the same size as vec
        - `n` is an optional argument in case one needs to compute partial differences
            `vec` and `signs` jointly  describe a vector of real numbers' sign and abs in log scale.
     Output:
        The first n-1 dimension of vec and signs will store the log-abs and sign of the difference.
     """
    #
    # And the first n-1 dimension of signs with the sign of the differences.
    # the sign is assigned to True to break symmetry if the diff is 0
    # Input:
    assert (vec.shape == signs.shape)
    if n < 0:
        n = np.max(vec.shape) - 1
    else:
        assert (np.max(vec.shape) >= n + 1)
    for j in range(0, n, 1):
        if signs[j] == signs[j + 1]:  # When the signs are the same
            # if the signs are both positive, then we can just use the standard one
            signs[j], vec[j] = stable_log_diff_exp(vec[j], vec[j + 1])
            # otherwise, we do that but toggle the sign
            if signs[j + 1] == False:
                signs[j] = ~signs[j]
        else:  # When the signs are different.
            vec[j] = stable_logsumexp_two(vec[j], vec[j + 1])
            signs[j] = signs[j + 1]

def stable_norm_ppf_one_minus_x(logx):
    """
    This function compute the normal inverse CDF at 1-x by taking logx as an input.
    Numerical stability has been taken into account with two levels of approximations.
    :param logx:
    :return: ppf(1-e^logx)
    """
    if logx < -30:
        # asymptotic approximation
        norm_ppf_one_minus_x = np.sqrt(-2 * logx - np.log(-2*logx) - np.log(2*np.pi))
    # # if logx < -10:
    # #     norm_ppf_one_minus_x = norm.ppf(-logx - logx ** 2 / 2 - logx ** 3 / 6)
    else:
        norm_ppf_one_minus_x = norm.ppf(1 - np.exp(logx))
    return norm_ppf_one_minus_x


def get_forward_diffs(fun, n):
    """ This is the key function for computing up to nth order forward difference evaluated at 0"""
    # Pre-compute the finite difference operators
    # Save them in log-scale
    func_vec = np.zeros(n + 3)  # _like(self.CGFs_int, float)
    signs_func_vec = np.ones(n + 3, dtype=bool)
    deltas = np.zeros(n + 2)  # ith coordinate of deltas stores log(abs(ith order discrete derivative))
    signs_deltas = np.zeros(n + 2, dtype=bool)
    for i in range(1, n + 3, 1):
        func_vec[i] = fun(1.0 * (i - 1))
    for i in range(0, n + 2, 1):
        # Diff in log scale
        #tmp = np.diff((signs_func_vec-0.5)*2 * np.exp(func_vec))

        stable_inplace_diff_in_log(func_vec, signs_func_vec, n=n + 2 - i)

        #tmp2 = (signs_func_vec-0.5)*2 *  np.exp(func_vec)

        deltas[i] = func_vec[0]
        signs_deltas[i] = signs_func_vec[0]
    return deltas, signs_deltas



def subsample_epsdelta(eps,delta,prob):
    """
    :param eps: privacy loss eps of the base mechanism
    :param delta: privacy loss delta of the base mechanism
    :param prob: subsampling probability
    :return: Amplified eps and delta
    This result applies to both subsampling with replacement and Poisson subsampling.
    The result for Poisson subsmapling is due to Theorem 1 of :
    Li, Ninghui, Qardaji, Wahbeh, and Su, Dong. On sampling, anonymization, and differential privacy or,
    k-anonymization meets differential privacy
    The result for Subsampling with replacement is due to:
    Jon Ullman's lecture notes: http://www.ccs.neu.edu/home/jullman/PrivacyS17/HW1sol.pdf
    See the proof of (b)
    """

    return np.log(1+prob*(np.exp(eps)-1)), prob*delta



def get_forward_diffs_naive(fun,n):
    func_vec = np.zeros(n + 3)  # _like(self.CGFs_int, float)
    signs_func_vec = np.ones(n + 3, dtype=bool)
    deltas = np.zeros(n + 2)  # ith coordinate of deltas stores log(abs(ith order discrete derivative))
    signs_deltas = np.zeros(n + 2, dtype=bool)
    for i in range(0, n + 3, 1):
        func_vec[i] = np.exp(fun(1.0 * (i - 1)))
    for i in range(0, n + 2, 1):
        # Diff in log scale
        #tmp = np.diff((signs_func_vec-0.5)*2 * np.exp(func_vec))

        func_vec=np.diff(func_vec)
        #tmp2 = (signs_func_vec-0.5)*2 *  np.exp(func_vec)

        deltas[i] = np.log(np.abs(func_vec[0]))
        signs_deltas[i] = (func_vec[0] >=0)
    return deltas, signs_deltas


def get_forward_diffs_direct(fun,n):
    func_vec = np.zeros(n + 3)  # _like(self.CGFs_int, float)
    signs_func_vec = np.ones(n + 3, dtype=bool)
    deltas = np.zeros(n + 2)  # ith coordinate of deltas stores log(abs(ith order discrete derivative))
    signs_deltas = np.zeros(n + 2, dtype=bool)
    signs_deltas_out = np.zeros(n + 2, dtype=bool)
    for i in range(0, n + 3, 1):
        func_vec[i] = fun(1.0 * (i - 1))

    C_stirling = np.zeros(n+3)
    anchor_point = np.zeros(n + 3)
    for i in range(0, n + 2, 1):
        # i+1 choose 0 to i+1 choose (i+1)/2
        coeff =  fun(1.0 * (i))/(i+1)
        func1 = lambda x: x * coeff + np.log(1 - np.exp(fun(x - 1) - x * coeff))

        for j in range(1,i+1,1):
            C_stirling[j] = np.log(comb(i+1,j))#logcomb(i+1,j)
        for j in range(1,i+1,1):
            anchor_point[j] = func1(j)
        tmp = anchor_point[0:(i+1)+1] + C_stirling[0:(i+1)+1]

        # Examples of these coefficients
        # -1, 1                   i+1= 1  odd
        # 1，-2， 1               i+1 = 2  even
        # -1，3，-3，1            i+1 = 3 odd
        # 1，-4，6，-4，1          i+1 = 4 even
        signs_deltas[i] = True
        deltas[i] = -np.inf
        for j in range(0,i+1,2):
            s, out = stable_log_diff_exp(tmp[j], tmp[j + 1])
            signs_deltas[i], deltas[i] = stable_sum_signed(signs_deltas[i], deltas[i], s, out)

        if not (i+1)%2: # even
            signs_deltas[i] = ~signs_deltas[i]
            signs_deltas[i], deltas[i] = stable_sum_signed(signs_deltas[i], deltas[i], False, tmp[i + 1])

        #
        # Lastly toggle the sign and add the anchor point back

        signs_deltas_out[i], deltas[i] = stable_sum_signed(~signs_deltas[i], deltas[i],
                                                     True, np.log(np.exp(coeff) - 1) * (i + 1))
    return deltas, signs_deltas_out


def logcomb(n, k):

    return (gammaln(n+1) - gammaln(n-k+1) - gammaln(k+1))

def get_binom_coeffs(sz):
    C = np.zeros(shape = (sz + 1, sz + 1));
    #for k in range(1,sz + 1,1):
    #    C[0, k] = -np.inf
    for n in range(sz + 1):
        C[n, 0] = 0  # 1
    for n in range(1,sz + 1,1):
        C[n, n] = 0
    for n in range(1,sz + 1,1):
        for k in range(1,n,1):
            # numerical stable way of implementing the recursion rule
            C[n, k] = stable_logsumexp_two(C[n - 1, k - 1],C[n - 1, k])
    # only the lower triangular part of the matrix matters
    return C

def get_binom_coeffs_dict(sz):
    C = {} #np.zeros(shape = (sz + 1, sz + 1));
    #for k in range(1,sz + 1,1):
    #    C[0, k] = -np.inf
    for n in range(sz + 1):
        C[(n, 0)] = 0  # 1
    for n in range(1,sz + 1,1):
        C[(n, n)] = 0
    for n in range(1,sz + 1,1):
        for k in range(1,n,1):
            # numerical stable way of implementing the recursion rule
            C[(n, k)] = stable_logsumexp_two(C[(n - 1, k - 1)],C[(n - 1, k)])
    # only the lower triangular part of the matrix matters
    return C

def expand_binom_coeffs_dict(C,sz,sznew):
    for n in range(sz,sznew+1,1):
        C[(n,0)] = 0
    for n in range(sz,sznew + 1,1):
        C[(n, n)] = 0
    for n in range(sz,sznew + 1,1):
        for k in range(1,n,1):
            C[(n, k)] = stable_logsumexp_two(C[(n - 1, k - 1)],C[(n - 1, k)])
    return C # also need to update the size of C to sznew whenever this function is called just to keep track.


def RDP_linear_interpolation(func,x):
    # linear interpolation upper bound through the convexity of CGF
    epsinf = func(np.inf)

    if np.isinf(x):
        return epsinf
    if (x >= 1.0) and (x <= 2.0):
        return np.minimum(epsinf, func(2.0))
    if np.equal(np.mod(x, 1), 0):
        return np.minimum(epsinf, func(x))
    xc = math.ceil(x)
    xf = math.floor(x)
    return np.minimum(
        epsinf,
        ((x - xf) * (xc-1)* func(xc) + (1 - (x - xf)) * (xf-1)*func(xf)) / (x - 1)
    )





# This module implements the log of characteristic  function templates for various
# mechanisms (see  https://arxiv.org/abs/2106.08567)
# We will use a pair of characteristic function to describe each mechanism.
# The conversion back to delta(eps) or eps(delta) will use numerical inversion, which details in converter.py


import numpy as np
from autodp import utils
import scipy.integrate as integrate
from autodp.utils import stable_logsumexp, _log1mexp






def phi_gaussian(params, t):
    """
    The closed-form log phi-function  of Gaussian mechanism.
    The log of the phi-function is (-1/2sigma^2(t^2 -it)).

    Args:
        t: the order of the  characteristic function.
        sigma: the std of the noise divide by the l2 sensitivity.

    Returns:
        The log of the phi-function evaluated at the t-th order.
    """
    sigma = params['sigma']
    result = -1.0 / (2 * sigma ** 2) * (t ** 2 - t * 1.0j)
    return result


def phi_rr_p(params, t):
    """
    The closed-form log phi-function  for Randomized Response, where
    the privacy loss R.V. is drawn from the distribution P.

    Args:
        t: the order of the characteristic function.
        params: contains two parameters: p and q.
        p: the probability to output one for dataset X.
        q: the probability to output one for dataset X'. default q = 1 - p.

    Return:
        The log phi-function of randomized response.
        """
    p = params['p']
    # generalized randomized response
    q = params['q']
    term_1 = np.log(p / q)
    term_2 = np.log((1 - p) / (1 - q))
    a = []
    left = np.log(p) + t * 1.0j * term_1
    right = np.log(1 - p) + 1.0j * t * term_2
    a.append(left)
    a.append(right)
    return stable_logsumexp(a)


def phi_rr_q(params, t):
    """
    The closed-form log phi-function for Randomized Response, where
    the privacy loss R.V. is drawn from the distribution Q
    (see Definition 12 in https://arxiv.org/pdf/2106.08567.pdf)

    The generalized randomize response can represent any pure-DP mechanisms.
    Args:
        t: the order of the characteristic function.
        params: contains two parameters: p and q.
        p: the probability to output one for dataset X.
        q: the probability to output one for dataset X'. default q = 1 - p.

    Return:
        The log phi-function of randomized response.
    """
    p = params['q']
    q = params['p']
    term_1 = np.log(p / q)
    term_2 = np.log((1 - p) / (1 - q))
    a = []
    left = np.log(p) + t * 1.0j * term_1
    right = np.log(1 - p) + 1.0j * t * term_2
    a.append(left)
    a.append(right)
    return stable_logsumexp(a)


def phi_laplace(params, t):
    """
    The closed-form log phi-function fof  Laplace mechanism.
    Args:
        t: the order of the characteristic function.
        b: the parameter for Laplace mechanism.

    Return:
        The log phi-function of Laplace mechanism.
    """
    b = params['b']

    term_1 = 1.j * t / b
    term_2 = -(1.j * t + 1) / b

    result = utils.stable_logsumexp_two(utils.stable_logsumexp_two(0, np.log(1. / (2 * t * 1.j + 1))) + term_1,
                                        np.log(2 * t * 1.j) -
                                        np.log(2 * t * 1.j + 1) + term_2)
    return result + np.log(0.5)


def phi_subsample_gaussian_p(params, t, remove_only=False):
    """
    Return the  log phi-function of the poisson subsample Gaussian mechanism (evaluated at the t-th order).
    phi(t) := E_p e^{t log(p)/log(q)}.
    For Poisson sampling, if  P: N(1, sigma^2) and Q:(0, sigma^2) dominate the base mechanism for both adding
     and removing neighboring, relationship, then after sampling with probability 'prob'
    new_p: (1-prob)*Q + prob*P; new_q: Q will dominates the sampled mechanism for remove only relationship.
    new_p: P; new_q: (1-prob)P + prob*Q will dominates the sampled mechanism for add only relationship.
    For details, please refer to Theorem 11 in https://arxiv.org/pdf/2106.08567.pdf

    The phi-function of the subsample Gaussian is not symmetric. This function implements the log phi-function
    phi(t) := E_p e^{t log(p)/log(q)} using Gaussian quadrature directly.
     Args:
        gamma: the sampling ratio.
        sigma: the std of the noise divide by the l2 sensitivity.
    """

    sigma = params['sigma']
    gamma = params['gamma']

    """
    The qua function (used for Double Quadrature method) computes e^{it log(p)/log(q)} 
    Gaussian quadrature requires the integration interval is [-1, 1]. To integrate over [-inf, inf], we 
    first convert y (the integral in Gaussian quadrature) to new_y.
    The privacy loss R.V. log(p/q) = log(gamma * e^(2 * new_y - 1)/2 * sigma**2 + 1 -gamma)
    """
    def qua(y):
        new_y = y * 1.0 / (1 - y ** 2)
        if remove_only is True:
            # stable computes log(new_p(new_y)/new_q(new_y))
            stable = utils.stable_logsumexp_two(np.log(gamma) + (2 * new_y-1) / (2 * sigma ** 2), np.log(1 - gamma))
            exp_term = np.exp(1.0j * stable * t)
            # density_term computes the pdf of new_p: (1-prob)*Q + prob*P
            density_term = utils.stable_logsumexp_two(- (new_y-1) ** 2 / (2 * sigma ** 2) + np.log(gamma), - (new_y)
                                                                ** 2 / (2 * sigma ** 2) + np.log(1-gamma))
            inte_function = np.exp(density_term) * exp_term


        else:
            # returns the add_only result.
            # new p(x): 1/sqrt{2pi*sigma**2}e^{-(x-1)^2/2sigma**2} new q(x) = (1-gamma)/(sqrt{2pi*sigma^2}*e^{-(x-1)**2
            # /2sigma**2} + gamma/sqrt{2pi*sigma**2}*e^{-(x)**2/2sigma**2}
            stable = -utils.stable_logsumexp_two(np.log(1-gamma), np.log(gamma)+(1-2*new_y)/(2*sigma**2))
            # Stable is for log new_P(x)/new_Q(x)
            exp_term = np.exp(1.0j * stable * t)
            density_term = -(new_y-1) ** 2 / (2 * sigma ** 2)
            inte_function = np.exp(density_term) * exp_term
        return inte_function

    # inte_f implements the integration over an infinite intervals.
    # int_-infty^infty f(x)dx = int_-1^1 f(y/(1-y^2)) * (1 + y**2) / ((1 - y ** 2) ** 2).
    inte_f = lambda y: qua(y) * (1 + y ** 2) / ((1 - y ** 2) ** 2)

    # Double quadrature: res computes the phi-function using Gaussian quadrature.
    res = integrate.quadrature(inte_f, -1.0, 1.0, tol=1e-15, rtol=1e-15, maxiter=100)
    result = np.log(res[0]) - np.log(np.sqrt(2 * np.pi) * sigma)
    return result


def phi_subsample_gaussian_q(params, t, remove_only=False):
    """
    Return the  log phi-function of the poisson subsample Gaussian mechanism (evaluated at the t-th order).
    phi(t) := E_q e^{t log(q)/log(p)}.

    For Poisson sampling, if  P: N(1, sigma^2) and Q:(0, sigma^2) dominate the base mechanism for both adding
     and removing neighboring, relationship, then after sampling with probability 'prob'
    new_p: (1-prob)*Q + prob*P; new_q: Q will dominates the sampled mechanism for remove only relationship.
    new_p: P; new_q: (1-prob)P + prob*Q will dominates the sampled mechanism for add only relationship.
    For details, please refer to Theorem 11 in https://arxiv.org/pdf/2106.08567.pdf

    Args:
        gamma: the sampling ratio.
        sigma: the std of the noise divide by the l2 sensitivity.

    Returns:
        The log phi-function evaluated at the t-th order.
    """

    sigma = params['sigma']
    gamma = params['gamma']

    """
    The qua function (used for Double Quadrature method) computes e^{it log(p)/log(q)}. 
    Gaussian quadrature requires the integration interval is [-1, 1]. To integrate over [-inf, inf], we 
    first convert y (the integral in Gaussian quadrature) to new_y.
    The privacy loss R.V. log(p/q) = -log(gamma * e^(2 * new_y - 1)/2 * sigma**2 + 1 -gamma)
    """

    def qua(y):
        new_y = y * 1.0 / (1 - y ** 2)
        if remove_only is True:

            log_q_p = -1.0 * utils.stable_logsumexp_two(np.log(gamma) + (2 * new_y - 1) / (2 * sigma ** 2),
                                                           np.log(1 - gamma))
            phi_result = np.exp(log_q_p * 1.0j * t)
            inte_function = phi_result * np.exp(-new_y ** 2 / (2 * sigma ** 2))
        else:
            # for add_only: log_q_p is log(new_Q(new_y)/new_P(new_y))
            log_q_p = utils.stable_logsumexp_two(np.log(1 - gamma), np.log(gamma) + (1 - 2 * new_y) / (2 * sigma ** 2))

            exp_term = np.exp(log_q_p * 1.0j *t)
            # density is (1-gamma)*P + gamma*Q, which is (1-gamma)*e^{-(x-1)^2/2sigma^2} + gamma * np.exp(-x^2/2sigma^2)
            density_term = utils.stable_logsumexp_two(-(new_y-1)**2/(2*sigma**2)+np.log(1-gamma),np.log(gamma)-(new_y)
                                                      **2/(2*sigma**2))
            inte_function = exp_term * np.exp(density_term)

        return inte_function

    # inte_f implements the integration over an infinite intervals.
    # int_-infty^infty f(x)dx = int_-1^1 f(y/1-y^2) * (1 + y**2) / ((1 - y ** 2) ** 2).
    inte_f = lambda y: qua(y) * (1 + y ** 2) / ((1 - y ** 2) ** 2)


    # Double quadrature: res computes the phi-function using Gaussian quadrature.
    res = integrate.quadrature(inte_f, -1.0, 1.0, tol=1e-15, rtol=1e-15, maxiter=100)
    result = np.log(res[0]) - np.log(np.sqrt(2 * np.pi) * sigma)
    return result





"""
A collection of analytical expressions of the Renyi Differential Privacy for popular randomized algorithms.
All expressions in this file takes in a function of the RDP order alpha and output the corresponding RDP.

These are used to create symbolic randomized functions for the RDP accountant to track.

Some of the functions contain the renyi divergence of two given distributions, these are useful to keep track of
the per-instance RDP associated with two given data sets.

TO CONTRIBUTORS:  1. any new addition to the rdp_bank should include a reference to the mechanism of
    interesting and the derivation of its RDP.
    2. You should try providing an implementation fo the entire range of alpha >0. If you do not
       have alpha <1, feel free to use any upper bound, e.g., the bound for alpha = 1.

"""


import numpy as np
import math
from autodp import utils
from autodp.utils import _log1mexp


def stable_log_diff_exp(x):
    # ensure that y > x
    # this function returns the stable version of log(exp(y)-exp(x)) if y > x

    mag = np.log(1 - np.exp(x - 0))

    return mag


def RDP_gaussian(params, alpha):
    """
    Args:
        sigma: normalized noise level: std divided by global L2 sensitivity
        alpha: The order of the Renyi Divergence

    Return: Evaluation of the RDP's epsilon
    """
    sigma = params['sigma']
    assert(sigma > 0)
    assert(alpha >= 0)
    return 0.5 / sigma ** 2 * alpha


def RDP_laplace(params, alpha):
    """
    Args:
        b: the ratio of the scale parameter and L1 sensitivity
        alpha: The order of the Renyi Divergence
    Return: Evaluation of the RDP's epsilon
    """

    b = params['b']
    # assert(b > 0)
    # assert(alpha >= 0)
    alpha=1.0*alpha

    if np.isinf(alpha):
        return 1/b
    elif alpha == 1:
        # KL-divergence
        return 1 / b + np.exp(-1 / b) - 1
    elif alpha > 1:  # alpha > 1
        return utils.stable_logsumexp_two((alpha-1.0) / b + np.log(alpha / (2.0 * alpha - 1)),
                                           -1.0*alpha / b + np.log((alpha-1.0) / (2.0 * alpha - 1)))/(alpha-1)
    elif alpha == 0.5:
        return -2*(-1.0/(2*b) + np.log(1 + 1.0/(2*b)))#   -2*np.log(np.exp(-1.0/(2*b))*(1+1.0/(2*b)))
    else:
        return np.log(alpha/(2.0*alpha-1)*np.exp((alpha-1.0)/b) + (alpha-1.0)/(2.0*alpha-1)*np.exp(-1.0*alpha/b))/(alpha-1)
        # Handling the case when alpha = 1/2?

def RDP_zCDP(params,alpha):
    """
    This function implements the RDP of (xi,rho)-zCDP mechanisms.
    Definition 11 of https://arxiv.org/pdf/1605.02065.pdf
    (Extended to  alpha > 0; may need to check for your mechanism)
    :param params: rho --- zCDP parameter
    xi ---- optional zCDP parameter
    :param alpha: The order of the Renyi-Divergence
    :return: the implied RDP at level alpha
    """
    rho = params['rho']
    if 'xi' in params.keys():
        xi = params['xi']
    else:
        xi = 0
    assert (alpha >= 0)
    return xi + rho*alpha

def RDP_independent_noisy_screen(params, alpha):

    """
    The data-independent RDP of ``Noisy Screening" (Theorem 7 in Private-kNN).

    The method is described in https://openaccess.thecvf.com/content_CVPR_2020/html/Zhu_Private-kNN_Practical_Differential_Privacy_for_Computer_Vision_CVPR_2020_paper.html)
    The exact data-independent bound requires searching a max_count from [k/c, k] to maximize RDP_noisy_screening for any alpha

    Args:
        params: contains three parameters. params['thresh'] is the threshold for noisy screening,
        k is the number of neighbors in Private-kNN, sigma is the noisy scale.

    Returns:
        The RDP of data-independent noisy screening.
    """
    threshold =params['thresh']
    k = params['k']
    sigma = params['sigma']
    import scipy.stats
    rdp = []

    for adjacent in [+1, -1]:
        for max_count in [int(k/10),threshold]:

            logp = scipy.stats.norm.logsf(threshold - max_count, scale=sigma)
            logq = scipy.stats.norm.logsf(threshold - max_count+adjacent, scale=sigma)
            log1q = _log1mexp(logq)
            log1p = _log1mexp(logp)
            if alpha == 1:
                return np.exp(logp)*(logp-logq) +(1-np.exp(logp))*(log1p-log1q)
            elif np.isinf(alpha):
                return np.abs(np.exp(logp-logq)*1.0)

            term1 = alpha*logp -(alpha-1)*logq
            term2 = alpha*log1p -(alpha-1)*log1q
            log_term = utils.stable_logsumexp_two(term1,term2)
            rdp.append(log_term)
    log_term = np.max(log_term)
    return 1.0*log_term /(alpha-1)

def RDP_noisy_screen(params, alpha):

    """
    return the data-dependent RDP of ``Noisy Screening" (Theorem 7 in Private-kNN).

    Args:
        params contains two parameters logp and logq. logp denotes the log of
         probability of P(max_vote + noise > Threshold).

    """
    logp =params['logp']
    logq = params['logq']
    log1q = _log1mexp(logq)
    log1p = _log1mexp(logp)
    if alpha == 1:
        return np.exp(logp)*(logp-logq) +(1-np.exp(logp))*(log1p-log1q)
    elif np.isinf(alpha):
        return np.abs(np.exp(logp-logq)*1.0)

    term1 = alpha*logp -(alpha-1)*logq
    term2 = alpha*log1p -(alpha-1)*log1q
    log_term = utils.stable_logsumexp_two(term1,term2)

    return 1.0*log_term /(alpha-1)

def RDP_inde_pate_gaussian(params, alpha):
    """
    Return the data-independent RDP of Noisy Aggregation (the global sensitivity is 2).
    The method is descriped in https://arxiv.org/abs/1802.08908

    Args:
        sigma: noisy scale added to the vote count.

    """
    sigma = params['sigma']
    return 1.0/sigma **2 *alpha

def RDP_depend_pate_gaussian(params, alpha):

    """
    Return the data-dependent RDP of GNMAX (proposed in PATE2)
    Bounds RDP from above of GNMax given an upper bound on q (Theorem 6).

    Args:
      logq: Natural logarithm of the probability of a non-argmax outcome.
      sigma: Standard deviation of Gaussian noise.
      orders: An array_like list of Renyi orders.

    Returns:
      Upper bound on RPD for all orders. A scalar if orders is a scalar.

    Raises:
      ValueError: If the input is malformed.
    """
    logq = params['logq']
    sigma = params['sigma']

    if alpha == 1:
        p = np.exp(logq)
        w = (2 * p - 1) * (logq - _log1mexp(logq))
        return w
    if logq > 0 or sigma < 0 or np.any(alpha < 1):  # not defined for alpha=1
        raise ValueError("Inputs are malformed.")

    if np.isneginf(logq):  # If the mechanism's output is fixed, it has 0-DP.
        print('isneginf', logq)
        if np.isscalar(alpha):
            return 0.
        else:
            return np.full_like(alpha, 0., dtype=np.float)

    variance = sigma ** 2

    # Use two different higher orders: mu_hi1 and mu_hi2 computed according to
    # Proposition 10.
    mu_hi2 = math.sqrt(variance * -logq)
    mu_hi1 = mu_hi2 + 1

    orders_vec = np.atleast_1d(alpha)

    ret = orders_vec / variance  # baseline: data-independent bound

    # Filter out entries where data-dependent bound does not apply.
    mask = np.logical_and(mu_hi1 > orders_vec, mu_hi2 > 1)

    rdp_hi1 = mu_hi1 / variance
    rdp_hi2 = mu_hi2 / variance

    log_a2 = (mu_hi2 - 1) * rdp_hi2

    # Make sure q is in the increasing wrt q range and A is positive.
    if (np.any(mask) and logq <= log_a2 - mu_hi2 *
            (math.log(1 + 1 / (mu_hi1 - 1)) + math.log(1 + 1 / (mu_hi2 - 1))) and
            -logq > rdp_hi2):
        # Use log1p(x) = log(1 + x) to avoid catastrophic cancellations when x ~ 0.
        log1q = _log1mexp(logq)  # log1q = log(1-q)
        log_a = (alpha - 1) * (
                log1q - _log1mexp((logq + rdp_hi2) * (1 - 1 / mu_hi2)))
        log_b = (alpha - 1) * (rdp_hi1 - logq / (mu_hi1 - 1))

        # Use logaddexp(x, y) = log(e^x + e^y) to avoid overflow for large x, y.
        log_s1 = utils.stable_logsumexp_two(log1q + log_a, logq + log_b)
        log_s = np.logaddexp(log1q + log_a, logq + log_b)
        ret[mask] = np.minimum(ret, log_s / (alpha - 1))[mask]
    # print('alpha ={} mask {}'.format(alpha,ret))
    if ret[mask] < 0:
        print('negative ret', ret)
        print('log_s1 ={} log_s = {}'.format(log_s1, log_s))
        print('alpha = {} mu_hi1 ={}'.format(alpha, mu_hi1))
        print('log1q = {} log_a = {} log_b={} log_s = {}'.format(log1q, log_a, log_b, log_s))
        ret[mask] = 1. / (sigma ** 2) * alpha
        # print('replace ret with', ret)
    assert np.all(ret >= 0)

    if np.isscalar(alpha):
        return np.asscalar(ret)
    else:
        return ret


def RDP_randresponse(params, alpha):
    """
    The RDP of the random response mechanism.

    Args:
        p: is the Bernoulli probability p of outputting the truth
        alpha: The order of the Renyi Divergence

    :return: Evaluation of the RDP's epsilon
    """

    p = params['p']
    assert((p >= 0) and (p <= 1))
    # assert(alpha >= 0)
    if p == 1 or p == 0:
        return np.inf
    if alpha == 1:
        return (2 * p - 1) * np.log(p / (1 - p))
    elif np.isinf(alpha):
        return np.abs(np.log((1.0*p/(1-p))))
    else:  # alpha > 1
        return utils.stable_logsumexp_two(alpha * np.log(p) + (1 - alpha) * np.log(1 - p),
                                           alpha * np.log(1 - p) + (1 - alpha) * np.log(p))/(alpha-1)



def RDP_expfamily(params, alpha):
    """
    The RDP of the exponential mechanism.
    The details can be found in Proposition 29 and Remark 30 of Wang, Balle, Kasiviswanathan (2018).
    Args:
        Delta: max distance of the natural parameters between two adjacent data sets in a certain norms.
        'L' 'B' are lambda functions. They are upper bounds of the local smoothness and local Lipschitzness
        of the log-partition function A, as a function of the radius of the local neighborhood in that norm.
        alpha: The order of the Renyi Divergence

    """
    Delta = params['Delta']
    L  = params['L'] # Local smoothness function that takes radius kappa as a input.
    B = params['B']  # Local Lipschitz function that takes radius kappa as a input.

    return np.minimum(alpha * L(alpha*Delta) * Delta ** 2,
                      (B((alpha-1)*Delta) + B(Delta)) * Delta)


def pRDP_diag_gaussian(params, alpha):
    """
    :param params:
        'mu1', 'mu2', 'sigma1', 'sigma2', they are all d-dimensional numpy arrays
    :param alpha: The order of the Renyi Divergence
    :return: Evaluation of the pRDP's epsilon

    See page 27 of http://mast.queensu.ca/~communications/Papers/gil-msc11.pdf for reference.
    """
    # the parameter is mu1, mu2, sigma1, sigma2
    # they are all d-dimensional numpy arrays
    # Everything can be generalized to general covariance, but
    # calculating A and checking the positive semidefiniteness of A is tricky

    mu1 = params['mu1']
    mu2 = params['mu2']
    sigma1 = params['sigma1']
    sigma2 = params['sigma2']

    def extrapolate(a, b):
        return alpha * a +(1 - alpha) * b

    #A = extrapolate(sigma1 ** (-1), sigma2 ** (-1))

    sigma = extrapolate(sigma2, sigma1)

    if not (sigma > 0).all():
        return np.inf
    else:
        #sigma = extrapolate(sigma2, sigma1)
        return (alpha / 2.0 * np.dot((mu1 - mu2),  (mu1 - mu2) / sigma) - 0.5 / (alpha-1) *
                (np.sum(np.log(sigma)) - extrapolate(np.sum(np.log(sigma2)), np.sum(np.log(sigma1)))))



def RDP_svt_laplace(params, alpha):
    """
    The RDP of Laplace-based SVT.
    Args:
        params['b']: the Laplace noise scale divide the sensitivity.
        params['k']: the SVT algorithm stops either k queries is achieved
        or the cut-off c is achieved.

    """
    b = params['b']
    k = params['k']  # the algorithm stops either k is achieved or c is achieved
    c = max(params['c'], 1)

    alpha = 1.0 * alpha
    if alpha <= 1:
        eps_1 = (1 / b + np.exp(-1 / b) - 1)
    elif np.isinf(alpha):
        eps_1 = 1 / b
    else:  # alpha > 1
        eps_1 = utils.stable_logsumexp_two((alpha - 1.0) / b + np.log(alpha / (2.0 * alpha - 1)),
                                           -1.0 * alpha / b + np.log((alpha - 1.0) / (2.0 * alpha - 1))) / (alpha - 1)

    eps_2 = 1 / b  # infinity rdp on nu
    c_log_n_c = c * np.log(k / c)
    tilde_eps = eps_2 * (c + 1)  # eps_infinity
    ret_rdp = min(c * eps_2 + eps_1, c_log_n_c * 1.0 / (alpha - 1) + eps_1 * (c + 1))
    ret_rdp = min(ret_rdp, 0.5 * alpha * tilde_eps ** 2)
    if np.isinf(alpha) or alpha == 1:
        return ret_rdp
    # The following is sinh-based method
    tilde_eps = eps_2 * (c + 1)
    cdp_bound = np.sinh(alpha * tilde_eps) - np.sinh((alpha - 1) * tilde_eps)
    cdp_bound = cdp_bound / np.sinh(tilde_eps)
    cdp_bound = 1.0 / (alpha - 1) * np.log(cdp_bound)
    return min(ret_rdp, cdp_bound)


def RDP_gaussian_svt_cgreater1(params, alpha):
    """
    The RDP of the gaussian-based SVT with the cut-off parameter c>1.

    Args:
        k:the maximum length before svt stops
        sigma: noise added to the threshold.
        c: the cut-off parameter in SVT.
    """
    sigma = params['sigma']
    c = max(params['c'], 1)
    k = params['k']  # the algorithm stops either k is achieved or c is achieved
    rdp_rho = 0.5 / (sigma ** 2) * alpha
    c_log_n_c = c * np.log(k / c)
    ret_rdp = c_log_n_c * 1.0 / (alpha - 1) + rdp_rho * (c + 1)
    return ret_rdp


def RDP_gaussian_svt_c1(params, alpha):
    """
    The RDP of the gaussian-based SVT with the cut-off parameter c=1.

    The detailed algorithm is described in Theorem 8 in
    https://papers.nips.cc/paper/2020/file/e9bf14a419d77534105016f5ec122d62-Paper.pdf

    A special case, where $k$ is not required, occurs when:
    1. Queries are non-negative, and have low-sensitivity (assumed),
    2. $\sigma_2 \geq \sqrt{3} \sigma_1$,
    3. $\gamma = 2$
    
    Calculates in the standard bound and the special case (Proposition 10) when applicable.

    params args:
        k (int, optional): the maximum length before svt stops (used in Theorem 8. -- general case)
        sigma (float): noise added to the threshold ($\sigma_1$ ($\sigma_{\\rho}$) in [ZW2020]).
        sigma_nu (float): noise added to each query ($\sigma_2$ ($\sigma_{\nu}$) in [ZW2020]).
            Defaults to \sqrt{3} \sigma_{1}
        Delta (float): global sensitivity for each query
        margin (float): query threshold parameter ($T$ in [ZW2020])
        gamma (float): positive multiplicative factor (default=2)
    """
    # API expects `sigma` to be defined for all gaussian mechanisms, let sigma=sigma_1:
    sigma_1 = params['sigma']
    # [ZW2020] denotes sigma_nu as sigma_2. By default, calibrate query noise to apply Prop. 10:
    sigma_2 = params.get('sigma_nu', np.sqrt(3) * sigma_1)
    # AutoDP uses sigma2 as a param, for consistency we support it
    sigma_2 = params.get('sigma2', sigma_2)
    margin = params.get('margin', 0.)
    Delta = params.get('Delta', 1.)
    gamma = params.get('gamma', 2.)

    rdp_rho = alpha * Delta**2 / (sigma_1 ** 2)
    # By Theorem 8, $\eps_{\nu}(\alpha)$ is upper bounded by 2 \Delta.
    rdp_nu = 2 * alpha * Delta**2 / (sigma_2 **2 )

    # If $k$ provided, compute the standard bound
    ret_rdp = np.inf
    if 'k' in params:
        k = params['k']
        ret_rdp = np.divide(np.log(k) , alpha - 1, where=alpha!=1) + rdp_rho + rdp_nu
    
    # Check if the conditions for the Propisition 10 are satisfied:
    if sigma_2 < np.sqrt(gamma + 1) * sigma_1:
        # Check if k is defined, if not, no bound can be applied correctly
        assert 'k' in params, "$k$ must be defined if Proposition 10 is not applicable."
        return ret_rdp

    # Proposition 10. Bound the stopping time by \E[\E[K | \rho=z]^{\gamma}], where K is a random stopping time
    stop_time_log_part = np.sqrt(1 + gamma)
    stop_time_log_part = stop_time_log_part * np.power(np.sqrt(2 * np.pi), gamma)
    stop_time_log_part = stop_time_log_part * ( np.power(( (margin * (1 + gamma))/sigma_1), gamma ) + 1)
    stop_time_log_part = np.log(stop_time_log_part)
    stop_time_exp_part = (gamma * margin**2) / (2 * sigma_1**2)

    # note exp(0) = 1, which is implicitly added
    moment_term = utils.stable_logsumexp_two(0, stop_time_log_part + stop_time_exp_part)
    moment_term = moment_term / (gamma * (alpha - 1))

    moment_based_bound = rdp_rho + rdp_nu + moment_term
    return min(moment_based_bound, ret_rdp)



def RDP_pureDP(params,alpha):
    """
    This function generically converts pure DP to Renyi DP.
    It implements Lemma 1.4 of Bun et al.'s CDP paper.
    With an additional cap at eps.

    :param params: pure DP parameter
    :param alpha: The order of the Renyi Divergence
    :return:Evaluation of the RDP's epsilon
    """
    eps = params['eps']
    # assert(eps>=0)
    # if alpha < 1:
    #     # Pure DP needs to have identical support, thus - log(q(p>0)) = 0.
    #     return 0
    # else:
    #     return np.minimum(eps,alpha*eps*eps/2)

    assert (alpha >= 0)
    if alpha == 1:
        # Calculate this by l'Hospital rule
        return eps * (math.cosh(eps) - 1) / math.sinh(eps)
    elif np.isinf(alpha):
        return eps
    elif alpha > 1:
        # in the proof of Lemma 4 of Bun et al. (2016)
        s, mag = utils.stable_log_diff_exp(utils.stable_log_sinh(alpha * eps),
                                           utils.stable_log_sinh((alpha - 1) * eps))
        return (mag - utils.stable_log_sinh(eps)) / (alpha - 1)
    else:
        return min(alpha * eps * eps / 2, eps * (math.cosh(eps) - 1) / math.sinh(eps))


def RDP_zCDP(params,alpha):
    """
    This function implements the RDP of (xi,rho)-zCDP mechanisms.
    Definition 11 of https://arxiv.org/pdf/1605.02065.pdf

    (Extended to  alpha > 0; may need to check for your mechanism)


    :param params: rho --- zCDP parameter
    xi ---- optional zCDP parameter
    :param alpha: The order of the Renyi-Divergence
    :return: the implied RDP at level alpha
    """
    rho = params['rho']
    if 'xi' in params.keys():
        xi = params['xi']
    else:
        xi = 0
    assert (alpha >= 0)
    return xi + rho*alpha


def RDP_truncatedCDP(params,alpha):
    """
    This function implements the RDP of (rho,w)-tCDP mechanisms.
    See Definition 1 of
    https://projects.iq.harvard.edu/files/privacytools/files/bun_mark_composable_.pdf

    (Extended to  alpha > 0; may need to check for your mechanism)

    :param params: rho, w
    :param alpha: The order of the Renyi-Divergence
    :return: the implied RDP at level alpha
    """
    rho = params['rho']
    w = params['w']
    assert (alpha >= 0)
    if alpha < w:
        return rho*alpha
    else:
        return np.inf



def RDP_subsampled_pureDP(params, alpha):
    """
    The special function for approximating the privacy amplification by subsampling for pure DP mechanism.
    1. This is more or less tight
    2. This applies for both Poisson subsampling (with add-remove DP definition)
       and Subsampling (with replace DP Definition)
    3. It evaluates in O(1)

    :param params: pure DP parameter, (optional) second order RDP, and alpha.
    :param alpha: The order of the Renyi Divergence
    :return: Evaluation of the RDP's epsilon

    """
    eps = params['eps']
    #eps2 = params['eps2'] # this parameter is optional, if unknown just use min(eps,2*eps*eps/2)
    if 'eps2' in params.keys():
        eps2= params['eps2']
    else:
        eps2 = np.minimum(eps,eps*eps)
    prob = params['prob']
    assert((prob<1) and (prob >= 0))
    assert(eps >= 0 and eps2 >=0)


    def rdp_int(x):
        if x == np.inf:
            return eps
        s, mag = utils.stable_log_diff_exp(eps,0)
        s, mag2 = utils.stable_log_diff_exp(eps2,0)

        s, mag3 = utils.stable_log_diff_exp(x*utils.stable_logsumexp_two(np.log(1-prob),np.log(prob)+eps),
                                            np.log(x) + np.log(prob) + mag)
        s, mag4 = utils.stable_log_diff_exp(mag3, np.log(1.0*x/2)+np.log(x-1)+2*np.log(prob)
                                            + np.log( np.exp(2*mag) - np.exp(np.min([mag,2*mag,mag2]))))

        return 1/(x-1)*mag4


        ## The following is the implementation of the second line.
        # if x <= 2:
        #     # Just return rdp 2
        #     return utils.stable_logsumexp([0, np.log(1.0*2/2)+np.log(2-1)+2*np.log(prob)
        #                                    + np.min([mag2,mag,2*mag])])
        # else:
        #     return 1/(x-1)*utils.stable_logsumexp([0, np.log(1.0*x/2)+np.log(x-1)+2*np.log(prob)
        #                                            + np.min([mag2,mag,2*mag]),
        #                                           3*np.log(prob) + 3*mag + np.log(x) + np.log(x-1)
        #                                           + np.log(x-2) - np.log(6) +
        #                                           (x-3) * utils.stable_logsumexp_two(np.log(1-prob),np.log(prob)+eps)])

    if alpha < 1:
        return 0
    else:
        return utils.RDP_linear_interpolation(rdp_int, alpha)


def pRDP_asymp_subsampled_gaussian(params, alpha):
    """
    :param params:
    :param alpha: The order of the Renyi Divergence
    :return: Evaluation of the pRDP's epsilon

    See Example 19 of Wang, Balle, Kasiviswanathan (2018)
    """
    sigma = params['sigma']
    prob = params['prob']
    assert((prob<1) and (prob >= 0))

    # The example where we have an approximately worst case asymptotic data set
    thresh = sigma**2/(prob*(1-prob)) + 1
    if alpha <= 1:
        return 0
    elif alpha >= thresh:
        return np.inf
    else:
        return (prob ** 2 / (2*sigma**2) * alpha * (thresh-1) / (thresh - alpha)
                + np.log((thresh-1)/thresh) / 2
                ) + np.log((thresh-1) / (thresh - alpha)) / 2 /(alpha-1)


def pRDP_asymp_subsampled_gaussian_best_case(params, alpha):
    """
    :param params:
    :param alpha: The order of the Renyi Divergence
    :return: Evaluation of the pRDP's epsilon

    See Example 20 of Wang, Balle, Kasiviswanathan (2018)
    """
    sigma = params['sigma']
    prob = params['prob']
    n = params['n']
    assert((prob<1) and (prob >= 0))

    # The example where we have an approximately best case data set
    return prob**2 / (2*sigma**2 + prob*(1-prob)*(n-1.0/n)/2) * alpha



def pRDP_expfamily(params, alpha):
    """
    :param params:
        'eta1', 'eta2' are the natural parameters of the exp family distributions.
        'A' is a lambda function handle the log partition. `A' needs to handle cases of infinity.
        'mu' is the mean of the sufficient statistics from distribution 1.
    :param alpha: The order of the Renyi Divergence
    :return: Evaluation of the pRDP's epsilon
    """
    # Theorem 1.2.19 of http://mast.queensu.ca/~communications/Papers/gil-msc11.pdf
    eta1 = params['eta1']
    eta2 = params['eta2']
    A = params['A']
    mu = params['mu'] # This is used only for calculating KL divergence
    # mu is also the gradient of A at eta1.


    assert(alpha >= 1)

    if alpha == 1:
        return np.dot(eta1-eta2, mu) + A(eta1) - A(eta2)

    def extrapolate(a, b):
        return alpha * a + (1 - alpha) * b

    tmp = A(extrapolate(eta1, eta2))
    if np.isinf(tmp) or np.isnan(tmp):
        return np.inf
    else: # alpha > 1
        return (A(extrapolate(eta1, eta2)) - extrapolate(A(eta1), A(eta2))) / (alpha - 1)




from setuptools import setup
import os

here = os.path.dirname(os.path.abspath(__file__))


with open("README.md", 'r') as f:
    long_description = f.read()

def _parse_requirements(path):
  """Parses requirements from file."""
  with open(os.path.join(here, path)) as f:
    deps = []
    for line in f:
      deps.append(line.rstrip())
    return deps

setup(
   name='autodp',
   version='0.2.3',
   description='Automating Differential Privacy Computation',
   license="Apache",
   long_description="The package helps researchers and developers to correctly use advanced methods in differential privacy and obtain provable DP guarantees. The core of the package is an analytical moments accountant that keeps track of Renyi Differential Privacy in analytical forms.",
   author='Yu-Xiang Wang',
   author_email='yuxiangw@cs.ucsb.edu',
   url='https://github.com/yuxiangw/autodp',
   download_url = 'https://github.com/yuxiangw/autodp/archive/refs/tags/v0.2.3.tar.gz',   
   keywords = ['Differential Privacy','Moments Accountant','Renyi Differential Privacy'],
   packages=['autodp'],  #same as name
   install_requires=[_parse_requirements('requirements.txt')], #external packages as dependencies
   classifiers=['Development Status :: 3 - Alpha',
   'Intended Audience :: Developers',
   'Intended Audience :: Science/Research',
   'Topic :: Scientific/Engineering :: Artificial Intelligence',
   'Topic :: Security',
   'License :: OSI Approved :: Apache Software License',
   'Programming Language :: Python :: 3',
   ],
)



"""
Core classes for autodp:

Mechanism --- A `mechanism' describes a randomized algorithm and its privacy properties.
                All `mechanism's (e.g., those in the `mechanism_zoo' module) inherit this class.

Transformer ---  A transformer takes one or a list of mechanism and outputs another mechanism
                All `transformer's (e.g., those in the `transformer_zoo' module, e.g., amplificaiton
                 by sampling, shuffling, and composition) inherit this class.

Calibrator --- A `calibrator' takes a mechanism with parameters (e.g. noise level) and automatically
                choose those parameters to achieve a prespecified privacy budget.
                All `calibrator's (e.g., the Analytical Gaussian Mechanism calibration, and others
                in `calibrator_zoo'inherit this class)

"""

import numpy as np
from autodp import converter

class Mechanism():
    """
     The base mechanism will use typically two functions to describe the mechanism

    # Attributes (actually functions as well):
    # 1: Approximate DP:   epsilon as a function of delta
    # 2. Renyi DP:   RDP epsilon as a function of \alpha
    # 3. Approximate RDP:  approximate RDP. RDP conditioning on a failure probability delta0.
    # 4. f-DP:  Type II error as a function of Type I error. You can get that from Approximate-DP
    #           or FDP directly.
    # 5. epsilon:  Pure DP bound.  If not infinity, then the mechanism satisfies pure DP.
    # 6. delta0:  Failure probability which documents the delta to use for approximate RDP
    #             in the case when there are no information available about the failure event.
    # 7. local_flag: Indicates whether the guarantees are intended to be for local differential privacy
    # 8. group_size: Integer measuring the granuality of DP.  Default is 1.
    # 9. replace_one: Flag indicating whether this is for add-remove definition of DP
    #                 or replace-one version of DP,  default is False

    # CDP and approximate-CDP will be subsumed in RDP bounds
    #
    # If we specify RDP only then it will propagate the RDP calculations to approximate-DP
    # and to f-DP
    # If we specify pure-DP only then it propagates to RDP,  Approximate-DP, f-DP and so on.
    # If we specify approximate-DP only, then it implies an approximate RDP bound with \delta_0.
    # If we specify f-DP only then it propagates to other specifications.
    # If we specify multiple calculations, then it will take the minimum of all of them
    #      in each category
    """


    def __init__(self):
        # Initialize everything with trivial (non-private) defaults
        def RenyiDP(alpha):
            return np.inf

        def approxRDP(delta, alpha):
            return np.inf

        def approxDP(delta, estimated_eps=None):
            return np.inf

        def cdf_p2q(x):
            return None
        def cdf_q2p(x):
            return None
        def pdf_p(x):
            return None
        def pdf_q(x):
            return None
        def log_phi_p2q(x):
            return None
        def log_phi_q2p(x):
            return None

        def approx_delta(eps):
            return 1

        def fDP(fpr):
            fnr = 0.0
            return fnr

        self.RenyiDP = RenyiDP
        self.approxRDP = approxRDP
        self.approxDP = approxDP
        self.approx_delta =approx_delta
        self.fDP = fDP
        #TODO need a dedicated Flag to indicate whether the dominating pair description is available for this mechanism;
        # and which mode we are using.
        self.cdf_p2q = cdf_p2q # cdf of the privacy loss random variable log(p/q), where p, q denotes the pdf of dominating pair.
        self.cdf_q2p = cdf_q2p
        self.pdf_p = pdf_p # pdf of the dominating pair.
        self.pdf_q = pdf_q
        self.log_phi_p2q = log_phi_p2q # log phi function of the privacy loss distribution (log (pdf_p)/(pdf_q))
        self.log_phi_q2p = log_phi_q2p
        self.exact_phi = True # whether admit a closed-form phi-function
        self.eps_pureDP = np.inf  # equivalent to RenyiDP(np.inf) and approxDP(0).

        self.delta0 = np.inf  # indicate the smallest allowable \delta0 in approxRDP that is not inf

        self.group_size = 1  # transformation that increases group size.

        self.neighboring = 'add_remove' # the default setting is 'add_remove', other choice includes 'add_only',
        # 'remove_only', 'replace_one'
        self.local_flag = False  # for the purpose of implementating local DP.
        #  We can convert localDP to curator DP by parallel composition and by shuffling.
        self.updated = False  # if updated, then when getting eps, we know that directly getting
        self.tbd_range = []

        # approxDP is the tightest possible.

    def get_approxDP(self, delta):
        # Output eps as a function of delta
        return self.approxDP(delta)

    def get_approx_delta(self, eps):
        # Output delta as a function of epsilon
        return self.approx_delta(eps)

    def get_approxRDP(self, delta, alpha):
        # Output eps as a function of delta and alpha
        return self.approxRDP(delta, alpha)

    def get_RDP(self, alpha):
        # Output RDP eps as a function of alpha
        return self.RenyiDP(alpha)

    def get_fDP(self, fpr):
        # Output false negative rate as a function of false positive rate
        return self.fDP(fpr)

    def get_cdf_p2q(self, x):
        # Output cdf as a function of log(p/q)
        return self.cdf_p2q(x)
    def get_cdf_q2p(self, x):
        # Output cdf as a function of log(q/p)
        return self.cdf_q2p(x)
    def get_pureDP(self):
        return self.eps_pureDP

    def get_eps(self, delta):
        # Get the smallest eps fo multiple calculations
        eps = [self.get_pureDP(), self.get_approxDP(delta)]
        # add get eps from RDP and get eps from approx RDP
        # and check the 'updated' flag. if updated, no need to do much
        return np.min(eps)

    def propagate_updates(self, func, type_of_update,
                          delta0=0,
                          BBGHS_conversion=True,
                          fDP_based_conversion=False, n_quad=700):

        # This function receives a new description of the mechanisms and updates all functions
        # based on what is new by calling converters.

        if type_of_update == 'pureDP':
            # function is one number
            eps = func
            self.eps_pureDP = np.minimum(eps, self.eps_pureDP)

            approxdp_new = converter.puredp_to_approxdp(eps)
            self.approxDP = converter.pointwise_minimum(approxdp_new, self.approxDP)
            rdp_new = converter.puredp_to_rdp(eps)
            self.RenyiDP = converter.pointwise_minimum(rdp_new, self.RenyiDP)
            fdp_new = converter.puredp_to_fdp(eps)
            self.fDP = converter.pointwise_maximum(fdp_new, self.fDP)

            self.approxRDP = converter.approxdp_func_to_approxrdp(self.approxDP)

            self.delta0 = 0  # the minimum non-trivial approximate RDP is now 0

            # lambda x: np.maximum(fdp_new(x),self.fDP(x))
        elif type_of_update == 'approxDP':
            # func will be a tuple of two numbers
            eps = func[0]
            delta = func[1]

            self.approxRDP = converter.pointwise_minimum_two_arguments(self.approxRDP,
                                          converter.approxdp_to_approxrdp(eps, delta))

            def approx_dp_func(delta1):
                if delta1 >= delta:
                    return eps
                else:
                    return np.inf

            self.approxDP = converter.pointwise_minimum(self.approxDP, approx_dp_func)
            self.fDP = converter.pointwise_maximum(self.fDP, converter.approxdp_to_fdp(eps, delta))

            self.delta0 = np.minimum(self.delta0, delta)
        elif type_of_update == 'approxDP_func':
            # func outputs eps as a function of delta
            # optional input delta0, telling us from where \epsilon becomes infinity

            self.delta0 = np.minimum(delta0, self.delta0)
            self.fDP = converter.pointwise_maximum(self.fDP, converter.approxdp_func_to_fdp(func))
            self.approxRDP = converter.pointwise_minimum_two_arguments(self.approxRDP,
                                 converter.approxdp_func_to_approxrdp(func))
            self.approxDP = converter.pointwise_minimum(self.approxDP, func)

        elif type_of_update == 'RDP':
            # function output RDP eps as a function of alpha
            self.RenyiDP = converter.pointwise_minimum(self.RenyiDP, func)
            if fDP_based_conversion:

                fdp_log, fdp_grad_log = converter.rdp_to_fdp_and_fdp_grad_log(func)

                self.fDP = converter.pointwise_maximum(self.fDP, converter.rdp_to_fdp(self.RenyiDP))

                # # --------- debugging code below -----------------
                #
                # def fdp_grad(x):
                #     return -np.exp(fdp_grad_log(np.log(x)))[0]
                #
                # def plot_fdp(x):
                #     grad = fdp_grad(x)
                #     y = self.fDP(x)
                #
                #     def tangent_line(u):
                #         return y + grad*(u-x)
                #
                #     import matplotlib.pyplot as plt
                #
                #     fpr_list, fnr_list = self.plot_fDP()
                #     plt.figure(1)
                #     plt.plot(fpr_list, fnr_list)
                #     plt.plot(fpr_list, tangent_line(fpr_list))
                #     plt.show()
                #
                # plot_fdp(0.01)
                # # ------------------------------------------------

                self.approxDP = converter.pointwise_minimum(self.approxDP,
                                                            converter.fdp_fdp_grad_to_approxdp(
                                                                fdp_log, fdp_grad_log,
                                                                log_flag=True))
                self.approx_delta = converter.pointwise_minimum(self.approx_delta,
                          converter.rdp_to_delta(self.RenyiDP, BBGHS_conversion=BBGHS_conversion))

                # self.approxDP = converter.pointwise_minimum(self.approxDP,
                #                                            converter.fdp_to_approxdp(self.fDP))
            else:
                self.approxDP = converter.pointwise_minimum(self.approxDP,
                         converter.rdp_to_approxdp(self.RenyiDP, BBGHS_conversion=BBGHS_conversion))
                self.approx_delta = converter.pointwise_minimum(self.approx_delta,
                          converter.rdp_to_delta(self.RenyiDP, BBGHS_conversion=BBGHS_conversion))
                self.fDP = converter.pointwise_maximum(self.fDP,
                                                       converter.approxdp_func_to_fdp(
                                                           self.approxDP))


        elif type_of_update == 'fDP':
            # f-DP,  input is Type I error or fpr, output is Type II error or fnr
            self.fDP = converter.pointwise_maximum(self.fDP, func)

            self.approxDP = converter.pointwise_minimum(self.approxDP,
                                                        converter.fdp_to_approxdp(func))
            self.approxRDP = converter.pointwise_minimum(self.approxRDP,
                                                         converter.approxdp_func_to_approxrdp(
                                                             self.approxDP))
        elif type_of_update == 'fDP_and_grad':
            # the input will be two functions
            fdp = func[0]
            fdp_grad = func[1]
            self.fdp = converter.pointwise_maximum(self.fDP, fdp)
            self.approxDP = converter.pointwise_minimum(self.approxDP,
                                                        converter.fdp_fdp_grad_to_approxdp(fdp,
                                                              fdp_grad,log_flag=False))
            self.approxRDP = converter.pointwise_minimum(self.approxRDP,
                                                         converter.approxdp_func_to_approxrdp(
                                                             self.approxDP))
        elif type_of_update == 'fDP_and_grad_log':
            # the input will be two functions
            fun1 = func[0]
            fun2 = func[1]
            fdp = lambda x: 1 - np.exp(fun1(np.log(x)))
            self.fDP = converter.pointwise_maximum(self.fDP, fdp)
            self.approxDP = converter.pointwise_minimum(self.approxDP,
                                                        converter.fdp_fdp_grad_to_approxdp(fun1,
                                                                                           fun2,
                                                                                           log_flag=True))
            self.approxRDP = converter.pointwise_minimum(self.approxRDP,
                                                         converter.approxdp_func_to_approxrdp(
                                                             self.approxDP))
        elif type_of_update == 'approxRDP':
            # func is a function of alpha and delta
            self.delta0 = np.minimum(delta0, self.delta0)
            self.approxRDP = converter.pointwise_minimum_two_arguments(self.approxRDP, func)
            # TODO: Write a function that converts approximateRDP to approximateDP

            # TODO: Write a function that converts approximateRDP to fDP.

        elif type_of_update == 'pdf':
            # func contains pdfs of the dominating pair
            self.pdf_p = func[0]
            self.pdf_q = func[1]

            # Convert pdf to phi-function using gaussian quadrature. if currant mechanism admits a closed-form phi,
            # then this step is unnecessary.
            # WARNING: this conversion can be numerically unstable, thus, we set exact_phi to be True as a default setting.
            if self.exact_phi == False:
                def log_phi_p2q(t): return converter.pdf_to_phi(self.pdf_p, self.pdf_q, t)
                def log_phi_q2p(t): return converter.pdf_to_phi(self.pdf_q, self.pdf_p, t)
                #log_phi_p, log_phi_q = converter.pdf_to_phi(self.pdf_p, self.pdf_q)
                self.log_phi_p2q = log_phi_p2q
                self.log_phi_q2p = log_phi_q2p
                # Apply Gaussian quadrature to do numerical inversion.
                # cdf_p2q denotes the cdf of privacy loss random variable log(p/q)
                # cdf_q2p denotes the cdf of log(q/p)
                cdf_p2q = lambda x: converter.phi_to_cdf(log_phi_p2q, x, n_quad=n_quad)
                cdf_q2p = lambda x: converter.phi_to_cdf(log_phi_q2p, x, n_quad=n_quad)

                self.approxDP = converter.pointwise_minimum(self.approxDP,
                                                            converter.cdf_to_approxdp(cdf_p2q, cdf_q2p))
                self.approx_delta = converter.pointwise_minimum(self.approx_delta,
                                                                converter.cdf_to_approxdelta(cdf_p2q, cdf_q2p))

        elif type_of_update == 'log_phi':
            # Update Analytical Fourier accountant with a new pair of log characteristic functions.

            log_phi_p2q = func[0]
            log_phi_q2p = func[1]
            self.exact_phi = True
            self.log_phi_p2q = log_phi_p2q
            self.log_phi_q2p = log_phi_q2p
            # Apply Gaussian quadrature to do numerical inversion.
            # cdf_p2q is the cdf of log(p/q), and cdf_q2p is the cdf for log(q/p).
            cdf_p2q = lambda x: converter.phi_to_cdf(log_phi_p2q, x, n_quad = n_quad)
            cdf_q2p = lambda x: converter.phi_to_cdf(log_phi_q2p, x, n_quad = n_quad)

            # Other approaches to convert phi function to cdf functions (not recommended)
            #cdf_p2q = lambda l: converter.cdf_approx_fft(log_phi_p, l)
            #cdf_q2p = lambda l: converter.cdf_approx_fft(log_phi_q, l)

            self.approxDP = converter.pointwise_minimum(self.approxDP,
                                                        converter.cdf_to_approxdp(cdf_p2q,cdf_q2p))
            self.approx_delta = converter.pointwise_minimum(self.approx_delta,
                                                            converter.cdf_to_approxdelta(cdf_p2q, cdf_q2p))
        elif type_of_update == 'cdf':
            # Analytical CDF: cdf_p2q is for log(p/q), cdf_q2p is for log(q/p).
            cdf_p2q = func[0]
            cdf_q2p = func[1]
            # propagate to ApproxDP
            self.approxDP = converter.pointwise_minimum(self.approxDP,
                                                        converter.cdf_to_approxdp(cdf_p2q, cdf_q2p))
            # propagate to CDF
            self.cdf_p2q = cdf_p2q
            self.cdf_q2p = cdf_q2p

            # TODO: convert CDF to discrete phi functions (FFT based approaches).
            """
            self.log_phi_p2q = converter.cdf_phi_p(cdf_p2q)
            self.log_phi_q2p = converter.cdf_phi_q(cdf_q2p)
            """
            # propagate to ApproxDelta
            self.approx_delta = converter.pointwise_minimum(self.approx_delta,
                                                            converter.cdf_to_approxdelta(cdf_p2q, cdf_q2p))

        elif type_of_update =='log_phi_adv':
            # Todo: Future work when the phi-function is parametrized by
            #  more than one parameter.
            """
            log_phi_p2q = func[0]
            log_phi_q2p = func[1]
            cdf_p2q = lambda x, t: converter.cdf_approx(log_phi_p2q, x, tbd=t)
            cdf_q2p = lambda x, t: converter.cdf_approx(log_phi_q2p, x, tbd=t)

            # first find t that optimize for one particular delta
            # we can set delta = 1e-5, and try out the tbd
            normal_equation = lambda tbd: converter.cdf_to_approxdp_tbd(cdf_p2q, cdf_q2p, tbd)
            n = 20
            clip = (self.tbd_range[1] - self.tbd_range[0])*1.0/n
            tbd_list = [self.tbd_range[0] + t*clip for t in range(n)]
            # select a delta randomly
            result_list = [normal_equation(tbd)(1e-5) for tbd in tbd_list]
            result_list = np.array(result_list)
            t = tbd_list[np.argmax(result_list)]

            self.approxDP = converter.cdf_to_approxdp_adv(cdf_p2q, cdf_q2p,t)
            #self.approxDP = converter.pointwise_minimum(self.approxDP, converter.cdf_to_approxdp_tbd(func,t))

            #self.approx_delta = converter.pointwise_minimum(self.approx_delta,
            #                                                results[0])
            """



    def set_all_representation(self, mech):
        """
        Set all representations from a mechanism object. This is useful when constructing a
        mechanism object using transformers then wanted to convert to a mechanism class definition.

        :param mech:  Input mechanism object.
        :return: None
        """
        # Need to make sure that the input is a mechanism object

        self.approxDP = mech.approxDP
        self.RenyiDP = mech.RenyiDP
        self.fDP = mech.fDP
        self.eps_pureDP = mech.eps_pureDP
        self.delta0 = mech.delta0

    # Plotting functions: returns lines for people to plot outside

    # Plot FNR against FPR --- hypothesis testing interpretation of DP
    def plot_fDP(self, length=101):
        fpr_list = np.linspace(0, 1, length)
        fnr_list = np.array([self.get_fDP(fpr) for fpr in fpr_list])
        return fpr_list, fnr_list

    # Plot RDP function
    def plot_RDP(self, alphamax=101, length=101):
        alpha_list = np.linspace(0, alphamax, length)
        RDP_list = np.array([self.get_RDP(alpha) for alpha in alpha_list])
        return alpha_list, RDP_list


class Transformer():
    """
    A transformer is a callable object that takes one or more mechanism as input and
    **transform** them into a new mechanism
    """

    def __init__(self):
        self.name = 'generic_transformer'
        self.unary_operator = False  # If true it takes one mechanism as an input,
        # otherwise it could take many, e.g., composition
        self.preprocessing = False  # Relevant if unary is true, it specifies whether the operation
        # is before or after the mechanism, e.g., amplification by sampling is before applying the
        # mechanism, "amplification by shuffling" is after applying the LDP mechanisms
        self.transform = lambda x: x

    def __call__(self, *args, **kwargs):
        return self.transform(*args, **kwargs)


class Calibrator():
    """
    A calibrator calibrates noise (or other parameters) meet a pre-scribed privacy budget
    """

    def __init__(self):
        self.name = 'generic_calibrator'

        self.eps_budget = np.inf
        self.delta_budget = 1.0

        self.obj_func = lambda x: 0

        self.calibrate = lambda x: x
        # Input privacy budget, a mechanism with params,  output a set of params that works
        # while minimizing the obj_func as much as possible

    def __call__(self, *args, **kwargs):
        return self.calibrate(*args, **kwargs)



"""
This file implements the classical DP accountant by Kairouz, Oh, Viswanath
"""


import numpy as np
from scipy.optimize import minimize_scalar

from autodp import utils

import math


class DP_acct:
    """
    A class that keeps track of (eps,delta) of all mechanisms that got run so far.
    Then it allows delta => epsilon, and epsilon => delta queries.
    """
    #
    #DPlosses = []

    def __init__(self,disp=False):
        self.DPlosses = []
        self.eps_state1 = 0
        self.eps_state2 = 0
        self.eps_state3 = 0
        self.delta_state = 0
        self.delta_state2 = 0
        self.disp=disp

    def update_DPlosses(self,eps,delta):
        self.DPlosses.append([eps,delta])
        self.eps_state1 += eps
        self.eps_state2 += (np.exp(eps) - 1) * eps / (np.exp(eps) + 1)
        self.eps_state3 += eps ** 2
        self.delta_state += np.log(1-delta)
        self.delta_state2 += delta
        #update the optimal DPlosses here?

    def get_eps_delta_naive(self):
        return self.eps_state1, self.delta_state2

    def get_minimum_possible_delta(self):
        return 1-np.exp(self.delta_state)


    def get_eps(self,delta):
        """ make use of KOV'15 to calculate the composition for hetereogeneous mechanisms"""
        # TODO: further improve upon this with Mutagh & Vadhan's approximation algorithm
        assert(delta >= 0)
        if delta>=1:
            return 0
        if delta is 0: # asking for pure DP
            if self.delta_state2 is 0:
                return self.delta_state2
            else:
                return np.inf
                #if self.disp:
                #    print("Pure DP is not feabile. Choose non-zero delta")
                #return -1

        # 1-delta =  (1-deltatilde) exp(state)
        deltatilde = 1 - np.exp(np.log(1-delta) -self.delta_state)
        if deltatilde <= 0:
            return np.inf
            #if self.disp:
            #    print("The chosen delta is not feasible. delta needs to be at least ", 1-np.exp(self.delta_state))
            #return -1

        eps1 = self.eps_state1
        eps2 = self.eps_state2 + (self.eps_state3 * 2 * np.log(np.exp(1) + self.eps_state3 ** 0.5 / deltatilde))**0.5
        eps3 = self.eps_state2 + (2*self.eps_state3*np.log(1 / deltatilde))**0.5

        return np.minimum(np.minimum(eps1,eps2),eps3)